# -*- coding: utf-8 -*-

import random
from typing import List
from PIL import Image, ImageDraw, ImageColor, ImageFont
default_palette = [
    '#B04759', '#E76161','#F99B7D',
    '#146C94', '#19A7CE', '#E893CF',
]
font_paths = [
    "STHeiti Light",
    "微软雅黑",
]
for font_path in font_paths:
    try:
        # 尝试加载字体文件
        zh_font = ImageFont.truetype(font_path, size=25)
        break  # 如果加载成功，跳出循环
    except IOError:
        continue  # 如果加载失败，继续尝试下一个字体

def draw_vocab_graph(model_name: str, charset_stats:dict, vocab_size:int, filename: str, width=100, height=100, cell_size=10, margin=40, palette=default_palette):
    total_chars = sum([s['total'] for s in charset_stats.values()])

    # 定义图像大小
    image_width = width * cell_size + 1 + margin * 2
    height = total_chars // width + 1
    image_height = height * cell_size + 1 + margin * 6

    # 创建新的空白图像
    image = Image.new("RGBA", (image_width, image_height), "#EEEEEE")

    # 获取图像的像素访问对象
    pixels = image.load()

    grid_color = (255,255,255, 40)

    # 根据map绘制栅格
    draw = ImageDraw.Draw(image)
    i = 0
    level = 0
    for name, stats in charset_stats.items():
        for j, m in enumerate(stats['map']):
            x = i % width
            y = i // width
            c = ImageColor.getrgb(palette[level])
            alpha = int(30 + m * 225)
            c = c[:3] + (alpha,)
            draw.rectangle((x * cell_size + 1 + margin,
                y * cell_size + 1 + margin,
                x * cell_size + (cell_size-1) + margin,
                y * cell_size + (cell_size-1) + margin
                ), fill=c)
            i += 1
        level += 1
    
    # 在图片左下角写入模型名称
    draw.text(
        (margin + 10, image_height - margin - 60),
        "[ {} ]".format(model_name),
        fill="#000000",
        align="right",
        font=ImageFont.truetype("Anonymous Pro", 30))
    # 在模型名称下方写入字表大小
    draw.text(
        (margin + 40, image_height - margin - 20),
        "vocab size: {:,} ".format(vocab_size),
        fill="#000000",
        align="right",
        font=ImageFont.truetype("Anonymous Pro", 20))

    # 在图片右下角写入字表统计信息
    stats_name = ""
    for name in charset_stats.keys():
        stats_name += "{}:\n".format(name)
    draw.text(
        (image_width - margin - 700, image_height - margin - 140),
        stats_name,
        fill="#000000",
        align="left",
        font=zh_font)
    
    stats_value = ""
    for s in charset_stats.values():
        stats_value += "{:4} / {:4}  ({:.2%})\n".format(s['known'], s['total'], float(s['known'])/s['total'])
    draw.text(
        (image_width - margin - 270, image_height - margin - 140),
        stats_value,
        fill="#000000",
        align="right",
        font=zh_font)

    # 保存图像
    image.save(filename)



from setuptools import setup, find_packages

setup(
    name='vocab-coverage',
    version='0.1',
    packages=find_packages(),
    install_requires=[
        'Pillow',
        'protobuf==3.20.0',
        'sentencepiece',
        'tiktoken',
        'torch',
        'transformers',
    ],
    extras_require={
        'generator': [
            'beautifulsoup4',
            'bs4',
            'pandas',
            'requests',
        ],
    },
    description='A Python package designed to perform coverage analysis on Chinese vocabulary for language models.',
    long_description_content_type="text/markdown",
    long_description=open('README.md').read(),
    author='Tao Wang',
    author_email='twang2218@gmail.com',
    url='https://github.com/twang2218/model-vocab-check',
    license='Apache License 2.0',
    classifiers=[
        'Development Status :: 3 - Alpha',
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Text Processing :: Linguistic",
    ],
    entry_points={
        'console_scripts': [
            'model-vocab-coverage = vocab_coverage.model:main',
        ],
    },
)




# -*- coding: utf-8 -*-

import argparse
import json
import os
import sys

from transformers import AutoTokenizer

if __name__ == "__main__":
    sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from vocab_coverage.draw import draw_vocab_graph

def model_check(charsets, model_name:str, output_dir:str, debug=False):
    print("检查模型 {} 的字表".format(model_name))
    try:
        tokenizer = AutoTokenizer.from_pretrained(model_name, trust_remote_code=True)
    except Exception as e:
        if "LLaMATokenizer" in e.args[0]:
            from transformers import LlamaTokenizer
            tokenizer = LlamaTokenizer.from_pretrained(model_name, trust_remote_code=True)
        elif "aquila" in e.args[0]:
            from flagai.data.tokenizer import Tokenizer
            name = 'aquila-7b'
            cache_dir = os.path.join('./model', name)
            tokenizer = Tokenizer.from_pretrained(name, cache_dir=cache_dir)
            tokenizer.cls_token_id = tokenizer.token_start_id
            tokenizer.sep_token_id = tokenizer.token_end_id
            tokenizer.unk_token_id = tokenizer.token_unk_id if hasattr(tokenizer, 'token_unk_id') else None
            tokenizer.pad_token_id = tokenizer.token_pad_id if hasattr(tokenizer, 'token_pad_id') else None
            tokenizer.mask_token_id = tokenizer.token_mask_id if hasattr(tokenizer, 'token_mask_id') else None
            tokenizer.vocab_size = tokenizer.num_tokens
        elif "OpenAI" in e.args[0]:
            import tiktoken
            name = model_name.split("/")[-1]
            tokenizer = tiktoken.encoding_for_model(name)
            tokenizer.vocab_size = tokenizer.n_vocab
            tokenizer.cls_token_id = tokenizer.encode_single_token('<|endoftext|>')
            if debug:
                print(tokenizer._special_tokens)
        else:
            print("加载模型 {} 失败：{}".format(model_name, e))
            return

    charset_stats = {
        name: {
            'known': 0,
            'total': len(chars),
            'chars': chars,
            'map': [0 for _ in range(len(chars))]
        } for name, chars in charsets.items()
    }

    if debug:
        print(tokenizer)
        print('vocab size:', tokenizer.vocab_size)
        if hasattr(tokenizer, 'cls_token_id'):
            print('[Special Token ID] => cls: {}, sep: {}, pad: {}, unk: {}, mask: {}'.format(
                tokenizer.cls_token_id if hasattr(tokenizer, 'cls_token_id') else None,
                tokenizer.sep_token_id if hasattr(tokenizer, 'sep_token_id') else None,
                tokenizer.pad_token_id if hasattr(tokenizer, 'pad_token_id') else None,
                tokenizer.unk_token_id if hasattr(tokenizer, 'unk_token_id') else None,
                tokenizer.mask_token_id if hasattr(tokenizer, 'mask_token_id') else None
            ))

    try:
        # 对于 ChatGLM 等模型，有特殊的头部token，需要特殊处理
        prefix_token = tokenizer.convert_tokens_to_ids('▁')
    except:
        prefix_token = tokenizer.cls_token_id

    for name, chars in charsets.items():
        for i, c in enumerate(chars):
            # 编码
            try:
                tokens_ids = tokenizer.encode(c, add_special_tokens=False)
            except Exception as e:
                if "add_special_tokens" in e.args[0]:
                    tokens_ids = tokenizer.encode(c)
                else:
                    print("编码字 {} 失败：{}".format(c, e))
                    continue

            # 编码预处理
            tn = type(tokenizer).__name__
            if len(tokens_ids) > 0 and tokens_ids[0] == prefix_token:
                # 对有头部特殊token的编码，去掉头部特殊token
                tokens_ids = tokens_ids[1:]

            # 验证编码
            if len(tokens_ids) > 1:
                c2 = tokenizer.decode(tokens_ids)
                if c != c2:
                    print("[{}] 编码字 {}({}) 失败：{} != {}".format(tn, c, tokens_ids, c, c2))
                    
            # 识字程度判断
            if len(tokens_ids) < 1 or (len(tokens_ids) == 1 and hasattr(tokenizer, 'unk_token_id') and tokens_ids[0] == tokenizer.unk_token_id):
                # 未识别的字
                charset_stats[name]['map'][i] = 0
            elif len(tokens_ids) == 1:
                # 完全识别的字
                charset_stats[name]['map'][i] = 1
                charset_stats[name]['known'] += 1
            else: # len(tokens_ids) > 1
                # 一定程度上识别的字，并不计数，只计算识别程度
                charset_stats[name]['map'][i] = 1.0/len(tokens_ids) # 识别程度
                if debug:
                    print("[{}] 汉字({})被拆分了，编码为{}".format(tn, c, tokens_ids))

    # 统计显示
    for name, stats in charset_stats.items():
        print("字表{}：{}/{} ({:.2%})".format(name, stats['known'], stats['total'], float(stats['known'])/stats['total']))

    # 生成字表图
    filename = model_name.replace("/", "_") + ".png"
    os.mkdir(output_dir) if not os.path.exists(output_dir) else None
    filename = os.path.join(output_dir, filename)
    draw_vocab_graph(model_name, charset_stats, tokenizer.vocab_size, filename, width=150)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--model_name", type=str, default="shibing624/text2vec-base-chinese")
    parser.add_argument("--charset_file", type=str, default="charset.json", help="字表文件")
    parser.add_argument("--output_dir", type=str, default="images", help="输出目录")
    parser.add_argument("--debug", action='store_true', help="是否打印调试信息")
    args = parser.parse_args()
    charsets = json.load(open(args.charset_file, 'r'))
    model_check(charsets, args.model_name, args.output_dir, args.debug)

if __name__ == "__main__":
    sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    main()



# -*- coding: utf-8 -*-

import argparse
import json
from bs4 import BeautifulSoup
import pandas as pd
import requests

def get_chinese_chars():
    urls = [
        'https://www.zwbk2009.com/index.php?title=%E9%80%9A%E7%94%A8%E8%A7%84%E8%8C%83%E6%B1%89%E5%AD%97%E8%A1%A8%EF%BC%88%E4%B8%80%E7%BA%A7%E5%AD%97%E8%A1%A8%EF%BC%89',
        'https://www.zwbk2009.com/index.php?title=%E9%80%9A%E7%94%A8%E8%A7%84%E8%8C%83%E6%B1%89%E5%AD%97%E8%A1%A8%EF%BC%88%E4%BA%8C%E7%BA%A7%E5%AD%97%E8%A1%A8%EF%BC%89',
        'https://www.zwbk2009.com/index.php?title=%E9%80%9A%E7%94%A8%E8%A7%84%E8%8C%83%E6%B1%89%E5%AD%97%E8%A1%A8%EF%BC%88%E4%B8%89%E7%BA%A7%E5%AD%97%E8%A1%A8%EF%BC%89',
    ]
    charset = []
    for url in urls:
        r = requests.get(url)
        soup = BeautifulSoup(r.text, 'html.parser')
        chinese_chars = soup.css.select('#mw-content-text div p a')
        chars = []
        for c in chinese_chars:
            title = c.get('title')
            if title is not None:
                title = title.replace('（页面不存在）', '')
                chars.append(title)
        charset.append(chars)
    return charset

def generate_unicode_chinese_chars():
    # 生成Unicode中日韩统一表意文字
    # 中日韩统一表意文字区段包含了20,992个汉字，编码范围为U+4E00–U+9FFF，Unicode 14.0起已完整填满。 
    chars = [chr(i) for i in range(0x4E00, 0x9FFF+1)]
    return chars

def generate_cnc_chars():
    # 获取《常用國字標準字體表》中常用字和次常用字字表
    urls = [
        'https://raw.githubusercontent.com/ButTaiwan/cjktables/master/taiwan/edu_standard_1.txt',
        'https://raw.githubusercontent.com/ButTaiwan/cjktables/master/taiwan/edu_standard_2.txt',
    ]
    charset = []
    for url in urls:
        chars = []
        r = requests.get(url)
        lines = r.text.split('\n')
        for l in lines:
            l = l.strip()
            if len(l) > 0:
                if l[0] == '#':
                    continue
                w = l.split('\t')
                if len(w) > 0:
                    chars.append(w[0])
        charset.append(chars)
    return charset

def generate_charsets(filename:str):
    charset = {}
    # 获取《通用规范汉字表》中一级、二级、三级字表'
    s1 = get_chinese_chars()
    charset['《通用规范汉字表》一级汉字'] = s1[0]
    charset['《通用规范汉字表》二级汉字'] = s1[1]
    charset['《通用规范汉字表》三级汉字'] = s1[2]
    for k, v in charset.items():
        print(f"{k}共有{len(v)}个汉字")

    base_chars = set()
    for cc in charset.values():
        base_chars.update(cc)
    print(f"《通用规范汉字表》共有{len(base_chars)}个汉字")
    s2 = generate_cnc_chars()

    extra_chars = []
    for c in s2[0]:
        if c not in base_chars:
            extra_chars.append(c)
    charset['《常用國字標準字體表》甲表(增)'] = extra_chars
    print(f"《常用國字標準字體表》甲表共有{len(s2[0])}个汉字，其中{len(extra_chars)}个汉字不在《通用规范汉字表》中")
    for c in extra_chars:
        base_chars.update(c)

    extra_chars = []
    for c in s2[1]:
        if c not in base_chars:
            extra_chars.append(c)
    charset['《常用國字標準字體表》乙表(增)'] = extra_chars
    print(f"《常用國字標準字體表》乙表共有{len(s2[1])}个汉字，其中{len(extra_chars)}个汉字不在《通用规范汉字表》中")
    for c in extra_chars:
        base_chars.update(c)

    chars = generate_unicode_chinese_chars()
    extra_chars = []
    for c in chars:
        if c not in base_chars:
            extra_chars.append(c)
    print(f"《Unicode中日韩统一表意文字》共有{len(chars)}个汉字，其中{len(extra_chars)}个汉字不在《通用规范汉字表》和《常用國字標準字體表》中")
    charset['《Unicode中日韩统一表意文字》(增)'] = extra_chars

    print(f"共有{sum([len(c) for c in charset.values()])}个汉字")

    # 保存到JSON文件
    with open(filename, 'w') as f:
        json.dump(charset, f, ensure_ascii=False, indent=4)

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--charset_file", type=str, default="charset.json", help="字表文件")
    args = parser.parse_args()
    generate_charsets(args.charset_file)